class ROW
inherit IS_WHITESPACE
creation make
feature
    tag_list : LINKED_LIST[TAG_HOLDER]
    context : CONTEXT
    text : STRING

    next_row, prev_row : ROW

    make is
    do
	!!tag_list.make
    end

    put_tag_list(l : like tag_list) is
    do
	tag_list := l
    end

    context_stack : LINKED_LIST[CONTEXT] is
    local
	c : CONTEXT
    once
	!!Result.make
	!!c.make
	Result.add_last(c)
    end

    tag_stack : LINKED_LIST[TAG_HOLDER] is
    once
	!!Result.make
    end

    add_tag(t : TAG) is
    local
	th : TAG_HOLDER
	c : CONTEXT
    do
	!!th.make(Current, t)
	if t.tag_start then
	    tag_stack.add_last(th)
	    !!c.make_from(context_stack.last)
	    if t.is_redundant(c) then
		th.set_redundant
	    else
		tag_list.add_last(th)
		t.execute(c)
		context_stack.add_last(c)
	    end
	else
	    th.partner_with(tag_stack.last)
	    tag_stack.remove_last
	    if not th.redundant then
		tag_list.add_last(th)
		context_stack.remove_last
	    end
	end
    end

    output is
    local
	it : ITERATOR_ON_LINKED_LIST[TAG_HOLDER]
    do
	!!it.make(tag_list)
	from it.start
	until it.is_off
	loop
	    it.item.output
	    it.next
	end
	io.put_string(text)
    end

    show_info is
	--for debugging
    local
	it : ITERATOR_ON_LINKED_LIST[TAG_HOLDER]
    do
	context.show_info
	io.put_string(", ")
	io.put_string("Tags: ")
	!!it.make(tag_list)
	from it.start
	until it.is_off
	loop
	    it.item.show_info
	    it.next
	end
	io.put_string(", ")
	io.put_string("Text: %"")
	escaped(text)
	io.put_string("%"")
	io.put_new_line
    end

    context_memo : MEMO[CONTEXT] is
    once
	!!Result
    end

    dump is
    local
	i : INTEGER
	special : BOOLEAN
	skip : BOOLEAN
    do
	from i := text.lower
	until i > text.upper
	loop
	    skip := false
	    if is_whitespace(text.item(i)) then
		if not context.tt_flag then
		    special := true
		    if context_memo.item /= Void
			    and then space_match(context, context_memo.item) then
			skip := true
		    else
			context.dump_ws
		    end
		else
		    context.dump_ws
		end
	    else
		special := false
		context.dump
		context_memo.clear
	    end
	    if not skip then
		io.put_string(": %"")
		if special then
		    io.put_character(' ')
		    context_memo.set_item(context)
		else
		    escaped_char(text.item(i))
		end
		io.put_string("%"")
		io.put_new_line
	    end
	    i := i + 1
	end
    end

    escaped_char(c : CHARACTER) is
    do
	inspect c
	when '%T' then
	    io.put_string("\t")
	when '%R' then
	    io.put_string("\r")
	when '%N' then
	    io.put_string("\n")
	else
	    io.put_character(c)
	end
    end

    escaped(s : STRING) is
    local
	i : INTEGER
    do
	from i := s.lower
	until i > s.upper
	loop
	    escaped_char(s @ i)
	    i := i + 1
	end
    end

    finish_row(s : STRING) is
    do
	text := s
	put_context(context_stack.last)
	--compress whitespace here
	if not context.tt_flag then
	    text := merge_whitespace(s)
	end
    end

    merge_whitespace(s : STRING) : STRING is
    local
	it : ITERATOR[CHARACTER]
	c, lastc : CHARACTER
    do
	!!Result.make(s.count)
	lastc := 'A' --any non-whitespace char will do
	it := s.get_new_iterator
	from it.start
	until it.is_off
	loop
	    c := it.item
	    if is_whitespace(lastc) and is_whitespace(c) then
	    else
		Result.add_last(c)
	    end
	    lastc := c
	    it.next
	end
    end

    put_text(s : STRING) is
    do
	text := s
    end

    put_context(c : CONTEXT) is
    do
	context := c
    end

    --following code is a bit convoluted
    --for efficiency, I split into cases so we never end up with
    --empty text (except possibly at the end)
    append_space_row(r, r2 : like Current) is
	--r is the next row
	--r2 is the row after r
    local
	it : ITERATOR[TAG_HOLDER]
    do
	--we'll take out the contents of r:
	--Current gets r's text (the whitespace)
	--and we prepend tags to next row's tag_list
	--the context does not need to be saved
	it := r.tag_list.get_new_iterator
	from it.start
	until it.is_off
	loop
	    it.item.put_row(r2)
	    it.next
	end
	r.tag_list.append_collection(r2.tag_list)
	r2.put_tag_list(r.tag_list)

	--compress whitespace if possible when we add it
	if not text.is_empty and then is_whitespace(text.last) then
	    if context.tt_flag then
		text.append(r.text)
	    end
	else
	    text.append(r.text)
	end
    end

    prepend_space_row(r : like Current) is
	--r is the previous rowa and it contains a single space
    local
	it : ITERATOR[TAG_HOLDER]
    do
	it := r.tag_list.get_new_iterator
	from it.start
	until it.is_off
	loop
	    it.item.put_row(Current)
	    it.next
	end
	r.tag_list.append_collection(tag_list)
	put_tag_list(r.tag_list)

	--compress whitespace if possible when we add it
	if not text.is_empty and then is_whitespace(text.first) then
	    if context.tt_flag then
		text.prepend(r.text)
	    end
	else
	    text.prepend(r.text)
	end
    end
end
