class SML
inherit TAG_SINGLETON
creation make
feature
    make is
    do
	!!program.make
	init_tag_table
    end

    parse is
    do
	!!current_row.make
	!!current_str.make_empty
	from io.read_character
	until io.end_of_input
	loop
	    inspect io.last_character
	    when '<' then
		if not current_str.is_empty then
		    finish_row
		end
		parse_tag
	    else
		add_char(io.last_character)
	    end
	    io.read_character
	end
	if not current_str.is_empty or
		not current_row.tag_list.is_empty then
	    finish_row
	end
    end

    redundancy_check is
    local
	it : ITERATOR[ROW]
	it2 : ITERATOR[TAG_HOLDER]
    do
	it := program.get_new_iterator
	from it.start
	until it.is_off
	loop
	    it2 := it.item.tag_list.get_new_iterator
	    from it2.start
	    until it2.is_off
	    loop
		if it2.item.redundant then
		    io.put_string("bug: redundancy remaining%N")
		end
		it2.next
	    end
	    it.next
	end
    end

    optimize is
    local
	sw : SIMPLIFY_WHITESPACE
	rt : REMOVE_TAGS
	pl : PL_SHORTCUT
    do
	--has to be done in this order
	--show_info
	--redundancy_check
	!!sw.make(program)
	sw.simplify_whitespace
	--show_info
	--redundancy_check
	!!rt.make(program)
	rt.remove_tags
	--by now, in a row,
	--there are no end tags after a start tag (since tags must be nested)
	--for each category except U there is at most one tag
	--(categories are colour, size, B, I, U, EM, S, TT)
	--an EM tag never follows a S tag

	--show_info
	--redundancy_check
	!!pl.make(program)
	pl.pl_shortcut
	--show_info
    end

    program : TWO_WAY_LINKED_LIST[ROW]

    current_row : ROW

    add_tag(t : TAG) is
    do
	current_row.add_tag(t)
    end

    --we keep a current_str here instead of going through current_row.text
    --every time we get a character
    current_str : STRING

    add_char(c : CHARACTER) is
    do
	current_str.add_last(c)
    end

    finish_row is
    do
	current_row.finish_row(current_str)
	program.add_last(current_row)
	!!current_row.make
	!!current_str.make_empty
    end

    parse_tag is
    local
	word : STRING
	done : BOOLEAN
	t : TAG
    do
	!!word.make(2)
	from io.read_character
	until io.last_character = '>'
	loop
	    word.add_last(io.last_character)
	    io.read_character
	end
	if not tag_table.has(word) then
	    io.put_string("unimplemented tag: ")
	    io.put_string(word)
	    io.put_string(", sorry!%N")
	    die_with_code(1)
	else
	    add_tag(tag_table.at(word))
	end
    end

    output is
    local
	it : ITERATOR[ROW]
    do
	it := program.get_new_iterator
	from it.start
	until it.is_off
	loop
	    it.item.output
	    it.next
	end
    end

    show_info is
	--for debugging
    local
	it : ITERATOR[ROW]
    do
	it := program.get_new_iterator
	from it.start
	until it.is_off
	loop
	    it.item.show_info
	    it.next
	end
    end

    dump is
	--for debugging
    local
	it : ITERATOR[ROW]
    do
	it := program.get_new_iterator
	from it.start
	until it.is_off
	loop
	    it.item.dump
	    it.next
	end
    end
end
